import { HAP, Logging, Service } from 'homebridge'
import * as HomeKitTypes from 'hap-nodejs/dist/lib/gen'

export type MIoTSpec = { name: string; siid: number, piid: number }
export type MIoTSpecs = MIoTSpec[]
export type MIoTSpecMapping = { [name: string]: MIoTSpec }
export type MIoTSpecGetQuery = MIoTSpec & { did: string }
export type MIoTSpecsGetQuery = MIoTSpecGetQuery[]
export type MIoTSpecSetQuery = MIoTSpec & { did: string; value: any }
export type MIoTSpecsSetQuery = MIoTSpecSetQuery[]
export type MIoTSpecsResponse<T> = MIoTSpec & { did: string; value: T }
export type MIoTSpecsResponseMapping = { [name: string]: MIoTSpecsResponse<any> }
export type MIoTSpecsResponseValueMapping = { [name: string]: any }
export type MIoTDeviceIdentify = { name: string; token: string; address: string; }

export type MIIOSpec = string
export type MIIOSpecMapping = { [name: string]: MIIOSpec }
export type MIIOSpecGetQuery = MIIOSpec[]
export type MIIOSpecResponse = (string | number)[]
export type MIIOSpecResponseValueMapping = { [name: string]: any }

export type DeviceConfig = { name: string; address: string; token: string; }
export type DeviceConfigs = DeviceConfig[]


export class DeviceInstance {
  public id: string
  public did: string
  public timeout: number

  // MIoT
  public miioCall<T extends any> (action: 'get_properties', specQuery: MIoTSpecsGetQuery): MIoTSpecsResponse<T>[]
  public miioCall<T extends any> (action: 'set_properties', specQuery: MIoTSpecsSetQuery): MIoTSpecsResponse<T>[]
  // MIIO
  public miioCall (action: 'get_prop', specQuery: MIIOSpecGetQuery): MIIOSpecResponse
  public miioCall (action: string, specQuery: (string | number)[]): any
}


type ValueOf<T> = T[keyof T]
export type HomeKitTypes = ValueOf<typeof HomeKitTypes.Generated>
export type MIoTDeviceProps = {
  identify: MIoTDeviceIdentify
  characteristicsService: Service
}
export type MIoTRegisterConfig = {
  name?: string,
  get: { formatter: (valueMapping: MIoTSpecsResponseValueMapping) => any }
  set?: { formatter: (value: any, previousProperty: MIoTSpecsResponseValueMapping) => string | number; property: string }
}
export type MIIORegisterConfig = {
  name?: string,
  get: { formatter: (valueMapping: MIoTSpecsResponseValueMapping) => any }
  set?: { formatter: (value: any, previousProperty: MIIOSpecResponseValueMapping) => (string | number)[]; property: string }
}
export class MIoTDevice {
  constructor (props: MIoTDeviceProps)

  // MIoT
  public addMIoTSpec (specs: MIoTSpecMapping): void
  public addMIoTCharacteristicListener(type: HomeKitTypes, config: MIoTRegisterConfig): void
  // MIIO
  public addMIIOSpec (specs: MIIOSpecMapping): void
  public addMIIOCharacteristicListener(type: HomeKitTypes, config: MIIORegisterConfig): void
}


export type Resolver = (...args: any[]) => void
export class Queue {
  public append(identify: MIoTDeviceIdentify, resolver: Resolver): void
  public resolve(identify: MIoTDeviceIdentify, value: any): void
}
export const SharedQueue: InstanceType<typeof Queue>


export class Device {
  public getInstance(identify: MIoTDeviceIdentify): Promise<DeviceInstance>
}
export const SharedDevice: InstanceType<typeof Device>


export class Foundation {
  public log: Logging
  public hap: HAP
}
export const SharedFoundation: InstanceType<typeof Foundation>
